# """
# 3D camera motion
#
# These functions allow Ren'Py to simulate 3D camera motion.  If you use this
# feature, in the first, set additional layers.
#
#    init -1 python:
#        config.layers = [ 'master', 'l1', 'l2', 'l3', 'transient', 'screens', 'overlay']
#
# Second, register layers which participate to a 3D motion as 3d layers.
#    init python:
#        register_3d_layer( ('master', 1000), ('l1', 300), ('l2', 500), ('l3', 700))
#
# Third, move a camera
#    label start:
#        scene bg
#        show A onlayer l1
#        show B onlayer l2
#        show B onlayer l3
#        'Start motion'
#        'It takes 1 second to move a camera to (0, 130, 0)'
#        $ camera_move(0, 130, 0, 0, 1)
#        'It takes 5 seconds to move a camera to (-70, 130, 1000)'
#        $ camera_move(-70, 130, 1000, 0, 5)
#        'A camera moves to (0, 0, 0) at the moment'
#        $ camera_move(0, 0, 0)
#        'It takes 1 second to rotate a camera to 360 degrees'
#        $ camera_move(0, 0, 0, 360, 1)
#        'reset motion'
#        $ camera_reset()
#        'end'
#
# You also use GUI viewer. Press shift+c and open Camera Viewer if config.developer is True.
# Transform Viewer(shift+T) also helps you.
#
# Notice that 3D camera motion and transforms to a layer can't be usable at the same time.
#
# """

init -1600 python:

    ##########################################################################
    # Camera functions

    # value of _3d_layers isn't included in rollback.
    # so I copy it many times.
    _3d_layers = {}
    _camera_x = 0
    _camera_y = 0
    _camera_z = 0
    _camera_r = 0
    _film_size = 50


    def register_3d_layer(*tuples):
        """
         :doc: camera

         Register layers as 3D layers and those distances from a camera. 3D
         layers is applied transforms to by a camera motion. If you want to
         change distances in game, register again.

         `tuples`
              These should be a tuple or tuples composed of a layer name and
              the distance from a camera.
         """
        global _3d_layers
        _3d_layers = {}
        for t in tuples:
            # [xanchor, yanchor, zoom, rotate, z]
            if t[1] == 0:
                t[1] = 1
            _3d_layers[t[0]] = [.5, .5, 1., float(t[1])]

    def release_3d_layer(layer):
        """
         :doc: camera

         Release 3D layers and the transforms of it.
         """
        global _3d_layers
        # renpy.show_layer_at(Transform(xalign=.5, yalign=.5, zoom=1.0, rotate=0, transform_anchor=True), layer)
        renpy.show_layer_at(Transform(xalign=.5, yalign=.5, zoom=1.0), layer)
        del _3d_layers[layer]

    def camera_reset():
        """
         :doc: camera

         Reset a camera motion.
         """
        global _3d_layers, _camera_x, _camera_y, _camera_z, _camera_r
        _camera_x = 0
        _camera_y = 0
        _camera_z = 0
        _camera_r = 0
        af_3d_layers = {}
        for l in sorted(_3d_layers.keys()):
            af_3d_layers[l] = [.5, .5, 1., _3d_layers[l][3]]
            # renpy.show_layer_at(Transform(xalign=.5, yalign=.5, zoom=1.0, rotate=0, transform_anchor=True), l)
            renpy.show_layer_at(Transform(xalign=.5, yalign=.5, zoom=1.0), l)
        _3d_layers = af_3d_layers

    def set_film_size(film_size):
        """
         :doc: camera

         Set the film_size of camera and reset a camera motion.

         `film_size`
              This is a number used as the film_size of a camera.
         """
        global _film_size
        _film_size = film_size
        camera_reset()

    def camera_move(x, y, z, rotate=0, during=0, warper='linear', subpixel=False):
        """
         :doc: camera

         Move the coordinate of a camera and apply transforms to all 3D layers.

         `x`
              the x coordinate of a camera
         `y`
              the y coordinate of a camera
         `z`
              the z coordinate of a camera
         `rotate`
              Defaul 0, the rotate of a camera
         `during`
              Default 0, this is the second times taken to move a camera.
         `warper`
              Default 'linear', this should be string and the name of a warper
              registered with ATL.
         `subpixel`
              Default False, if True, causes things to be drawn on the screen
              using subpixel positioning
         """
        global _3d_layers, _camera_x, _camera_y, _camera_z, _camera_r, _film_size

        # adjustment x, y
        x = x/100.
        y = y/100.
        x0 = _camera_x/100.
        y0 = _camera_y/100.
        z0 = _camera_z
        r0 = _camera_r

        af_3d_layers = {}
        for l in sorted(_3d_layers.keys()):
            ax0 = _3d_layers[l][0]
            ay0 = _3d_layers[l][1]
            zoom0 = _3d_layers[l][2]
            l_z = _3d_layers[l][3]

            distance0 = float(_camera_z + l_z)
            distance = float(z + l_z)

            # time where a camera crosses subject
            cross_time = -1.
            if (distance < 0 and distance0 > 0) or (distance > 0 and distance0 < 0):
                cross_time = -distance0/(distance - distance0)
            elif distance < 0 and distance0 < 0:
                cross_time = 0

            # avoid zero division
            if distance == 0:
                distance = 0.1

            ax = ( _film_size*(x-x0) + distance0*(ax0-0.5) )/distance + 0.5
            ay = ( _film_size*(y-y0) + distance0*(ay0-0.5) )/distance + 0.5
            zoom  = (distance0/distance)*zoom0

            if during > 0:
                renpy.show_layer_at(Transform(function=renpy.curry(_camera_trans)(ax0=ax0, ax=ax, ay0=ay0, ay=ay, distance0=distance0, x0=x0, x=x, y0=y0, y=y, z0=z0, z=z, zoom0=zoom0, zoom=zoom, r0=r0, r=rotate, during=during, warper=warper, cross_time=cross_time, subpixel=subpixel)), l)
            else:
                if zoom >= 0:
                    # renpy.show_layer_at(Transform(xpos=.5, ypos=.5, xanchor=ax, yanchor=ay, zoom=zoom, rotate=r, transform_anchor=True, alpha=1), l)
                    renpy.show_layer_at(Transform(xpos=.5, ypos=.5, xanchor=ax, yanchor=ay, zoom=zoom, rotate=rotate, alpha=1, subpixel=subpixel), l)
                else:
                    # renpy.show_layer_at(Transform(xpos=.5, ypos=.5, xanchor=ax, yanchor=ay, zoom=zoom, rotate=r, transform_anchor=True, alpha=0), l)
                    renpy.show_layer_at(Transform(xpos=.5, ypos=.5, xanchor=ax, yanchor=ay, zoom=zoom, rotate=rotate, alpha=0, subpixel=subpixel), l)

            af_3d_layers[l] = [ax, ay, zoom, l_z]

        _camera_x = int(x*100)
        _camera_y = int(y*100)
        _camera_z = z
        _camera_r = rotate
        _3d_layers = af_3d_layers

    def _camera_trans(tran, st, at, ax0, ax, ay0, ay, distance0, x0, x, y0, y, z0, z, zoom0, zoom, r0, r, during, warper, cross_time, subpixel):
        g = renpy.atl.warpers[warper](st/during) 
        tran.subpixel = subpixel
        if g <= 1:
            distance = g*(z-z0)+distance0
            if distance == 0:
                distance = .1
            tran.xpos    = .5 
            tran.ypos    = .5

            tran.xanchor = ( _film_size*g*(x-x0) + distance0*(ax0-0.5) )/distance + 0.5
            tran.yanchor = ( _film_size*g*(y-y0) + distance0*(ay0-0.5) )/distance + 0.5
            # tran.xanchor = _film_size*((g*(x-x0)+x0)/distance - x0/distance0) + ax0
            # tran.yanchor = _film_size*((g*(y-y0)+y0)/distance - y0/distance0) + ay0

            # tran.transform_anchor = True
            tran.rotate  = g*(r  -  r0) + r0

            if zoom0 > 0 and zoom < 0:
                if cross_time <= g:
                    tran.alpha = 0
                else:
                    tran.alpha = 1
                    tran.zoom    = (distance0 / distance)*zoom0
            elif zoom0 < 0 and zoom > 0:
                if cross_time <= g:
                    tran.alpha = 1
                    tran.zoom    = (distance0 / distance)*zoom0
                else:
                    tran.alpha = 0
            elif zoom0 < 0:
                tran.alpha = 0
            elif zoom0 > 0:
                tran.alpha = 1
                tran.zoom    = (distance0 / distance)*zoom0
            return .005
        else:
            # tran.transform_anchor = True
            tran.xpos    = .5 
            tran.ypos    = .5
            tran.xanchor = ax
            tran.yanchor = ay
            tran.zoom    = zoom
            tran.rotate    = r
            if zoom >= 0:
                tran.alpha = 1
            else:
                tran.alpha = 0
            return None


    class CameraViewer(object):
        def __init__(self, range_x=5000, range_y=5000, range_z=10000, range_r=360, range_film_size=100, range_layer_z=10000):

            self.range_x         = range_x
            self.range_y         = range_y
            self.range_z         = range_z
            self.range_r         = range_r
            self.range_film_size = range_film_size
            self.range_layer_z   = range_layer_z

        def viewer(self):
            global _3d_layers, _film_size, _film_size_org
            if not config.developer:
                return
            self._3d_layers_org = _3d_layers.copy()
            _film_size_org = _film_size
            renpy.show_screen("_camera_viewer", viewer=self)
            renpy.restart_interaction()
            # renpy.call_in_new_context("_camera_viewer", viewer=self)

        def reset(self):
            global _3d_layers
            _3d_layers = self._3d_layers_org.copy()
            set_film_size(_film_size_org)
            renpy.restart_interaction()

        def x_changed(self, v):
            camera_move(v - self.range_x, _camera_y, _camera_z, _camera_r)
            renpy.restart_interaction()

        def y_changed(self, v):
            camera_move(_camera_x, v - self.range_y, _camera_z, _camera_r)
            renpy.restart_interaction()

        def z_changed(self, v):
            camera_move(_camera_x, _camera_y, v - self.range_z, _camera_r)
            renpy.restart_interaction()

        def r_changed(self, v):
            camera_move(_camera_x, _camera_y, _camera_z, v - self.range_r)
            renpy.restart_interaction()

        def f_changed(self, v):
            set_film_size(v)
            renpy.restart_interaction()
            
        def generate_layer_z_changed(self, l):
            def layer_z_changed(v):
                if v == 0:
                    v = 1
                _3d_layers[l][3] = float(v)
                camera_reset()
                renpy.restart_interaction()
            return layer_z_changed

        def put_clipboard(self, camera_tab):
            string = ""
            if camera_tab:
                string = '$camera_move'+str((_camera_x, _camera_y, _camera_z, _camera_r))
            else:
                for layer in sorted(_3d_layers.keys()):
                    string += "('"+layer+"', "+str(_3d_layers[layer][3])+"), "
                string = '$register_3d_layer('+ string[:-2] + ")"
            renpy.put_clipboard(string)
            renpy.notify('Putted "'+string+'" on clipboard')

# label _camera_viewer(viewer):
#     show screen _camera_viewer(viewer)
#     $ui.interact()
#
#     return

screen _camera_viewer(viewer, camera_tab=True):
    # key "game_menu" action Return()
    key "game_menu" action Hide("_camera_viewer")

    frame:
        yalign 1.
        style_group "gui_viewer_frame"
        has vbox
        hbox:
            xfill False
            textbutton _("Camera Setting") action [SelectedIf(camera_tab), Show("_camera_viewer", viewer=viewer)]
            textbutton _("3D Layers Setting") action [SelectedIf(not camera_tab), Show("_camera_viewer", viewer=viewer, camera_tab=False)]
        if camera_tab:
            hbox:
                label "x"
                label "[_camera_x:>7]"
                null width 10
                bar adjustment ui.adjustment(range=viewer.range_x*2, value=_camera_x+viewer.range_x, page=1, changed=viewer.x_changed) xalign 1.
            hbox:
                label "y"
                label "[_camera_y:>7]"
                null width 10
                bar adjustment ui.adjustment(range=viewer.range_y*2, value=_camera_y+viewer.range_y, page=1, changed=viewer.y_changed) xalign 1.
            hbox:
                label "z"
                label "[_camera_z:>7]"
                null width 10
                bar adjustment ui.adjustment(range=viewer.range_z*2, value=_camera_z+viewer.range_z, page=1, changed=viewer.z_changed) xalign 1.
            hbox:
                label "rotate"
                label "[_camera_r:>7]"
                null width 10
                bar adjustment ui.adjustment(range=viewer.range_r*2, value=_camera_r+viewer.range_r, page=1, changed=viewer.r_changed) xalign 1.
            hbox:
                label "film size"
                label "[_film_size:>7]"
                null width 10
                bar adjustment ui.adjustment(range=viewer.range_film_size, value=_film_size, page=1, changed=viewer.f_changed) xalign 1.
        else:
            for l in sorted(_3d_layers.keys()):
                hbox:
                    label "[l:<7]"
                    label "{:>7d}".format(int(_3d_layers[l][3]))
                    null width 10
                    bar adjustment ui.adjustment(range=viewer.range_layer_z, value=_3d_layers[l][3], page=1, changed=viewer.generate_layer_z_changed(l)) xalign 1.
        hbox:
            xfill False
            xalign .9
            textbutton _("clip board") action Function(viewer.put_clipboard, camera_tab)
            textbutton _("reset") action [viewer.reset, renpy.restart_interaction]
            textbutton _("close") action Hide("_camera_viewer")


##########################################################################
# TransformViewer

init -1600 python:
    class TransformViewer(object):
        def __init__(self):

            self.int_range = 1500
            self.float_range = 7.0
            # layer->tag->property->value
            self.state_org = {}
            # {property:default, float}, default is used when property can't be got.
            self.props_list = (
            "xpos",
            "ypos",
            "xanchor",
            "yanchor",
            "xzoom",
            "yzoom",
            "zoom",
            "rotate",
            "alpha",
            "additive"
            )
            self.props_default = {
            "xpos":(0., False),
            "ypos":(0., False),
            "xanchor":(0., False),
            "yanchor":(0., False),
            "xzoom":(1., True),
            "yzoom":(1., True),
            "zoom":(1., True),
            "rotate":(0, False),
            "alpha":(1., True),
            "additive":(0., True),
            }

        def viewer(self):
            if not config.developer:
                return
            sle = renpy.game.context().scene_lists
            # back up for reset()
            for layer in config.layers:
                self.state_org[layer] = {}
                for tag in sle.layers[layer]:
                    d = sle.get_displayable_by_tag(layer, tag[0])
                    pos = renpy.get_placement(d)
                    state = getattr(d, "state", None)

                    self.state_org[layer][tag[0]] = {}
                    for p in ["xpos", "ypos", "xanchor", "yanchor"]:
                        self.state_org[layer][tag[0]][p] = getattr(pos, p, None)
                    for p in self.props_list:
                        if not self.state_org[layer][tag[0]].has_key(p):
                            self.state_org[layer][tag[0]][p] = getattr(state, p, None)
            renpy.show_screen("_transform_viewer", viewer=self)
            renpy.restart_interaction()
            # renpy.call_in_new_context("_transform_viewer", viewer=self)

        def reset(self):
            for layer in config.layers:
                if layer in ['transient', 'screens', 'overlay']:
                    break
                for tag, props in self.state_org[layer].iteritems():
                    if tag and props:
                        kwargs = props.copy()
                        for p in self.props_list:
                            if kwargs[p] is None:
                                kwargs[p] = self.props_default[p][0]
                            if self.state_org[layer][tag]["rotate"] is None:
                                kwargs["rotate"] = None
                        renpy.show(tag, [Transform(**kwargs)], layer=layer)
            renpy.restart_interaction()

        def generate_changed(self, tag, prop, layer, int=False):
            def changed(v):
                kwargs = {}
                for p in self.props_list:
                    kwargs[p] = self.get_state(layer, tag, p, False)

                if int and not self.props_default[prop][1]:
                    kwargs[prop] = v - self.int_range
                else:
                    kwargs[prop] = v -self.float_range
                renpy.show(tag, [Transform(**kwargs)], layer=layer)
                renpy.restart_interaction()
            return changed

        def get_state(self, layer, tag, p, default=True):
            sle = renpy.game.context().scene_lists

            if tag:
                d = sle.get_displayable_by_tag(layer, tag)
                pos = renpy.get_placement(sle.get_displayable_by_tag(layer, tag))
                state = getattr(pos, p, None)
                if state is None:
                    state = getattr(getattr(d, "state", None), p, None)
                # set default
                if state is None and default:
                    state = self.props_default[p][0]
                if state and self.props_default[p][1]:
                    state = float(state)
                return state
            return None

        def put_clipboard(self, prop, value):
            renpy.put_clipboard(prop+" "+str(value))
            renpy.notify('Putted "'+prop+' '+str(value)+'" on clipboard')

# label _transform_viewer(viewer):
#     show screen _transform_viewer(viewer)
#     $ui.interact()
#
#     return

screen _transform_viewer(viewer, layer="master", tag=""):
    # key "game_menu" action Return()
    key "game_menu" action Hide("_transform_viewer")
    zorder 10

    frame:
        xfill True
        style_group "gui_viewer"
        vbox:

            hbox:
                hbox:
                    xfill False
                    label _("layers")
                    for l in config.layers:
                        if l not in ["screens", "transient", 'overlay', "debug"]:
                            textbutton "[l]" action [SelectedIf(l == layer), Show("_transform_viewer", viewer=viewer, layer=l)]
                hbox:
                    xfill False
                    xalign .9
                    textbutton _("reset") action [viewer.reset, renpy.restart_interaction]
                    textbutton _("close") action Hide("_transform_viewer")

            hbox:
                xfill False
                label _("images")
                for t in renpy.game.context().scene_lists.layers[layer]:
                    textbutton "[t[0]]" action [SelectedIf(t[0] == tag), Show("_transform_viewer", viewer=viewer, layer=layer, tag=t[0])]

            if tag:
                for p in viewer.props_list:
                    $state = viewer.get_state(layer, tag, p)
                    if isinstance(state, int):
                        hbox:
                            textbutton "[p]" action Function(viewer.put_clipboard, p, state)
                            label "[state:>10]"
                            null width 10
                            bar adjustment ui.adjustment(range=viewer.int_range*2, value=state+viewer.int_range, page=1, changed=viewer.generate_changed(tag, p, layer, True)) xalign 1.
                    elif isinstance(state, float):
                        hbox:
                            textbutton "[p]" action Function(viewer.put_clipboard, p, state)
                            label "[state:>10.4f]"
                            null width 10
                            bar adjustment ui.adjustment(range=viewer.float_range*2, value=state+viewer.float_range, page=.05, changed=viewer.generate_changed(tag, p, layer)) xalign 1.

init -1600:
    style gui_viewer_frame background "#0004"
    style gui_viewer_button size_group "gui_viewer"
    style gui_viewer_button_text xalign .5
    style gui_viewer_label xminimum 100
    style gui_viewer_vbox xfill True
    style gui_viewer_hbox xfill True
    style gui_viewer_frame_frame background "#0004"
    style gui_viewer_frame_label xminimum 100
    style gui_viewer_frame_hbox xfill True

init -1600 python:
    def _scene(layer='master'):

        renpy.scene(layer)
        # Also clear 3d layers
        for l in _3d_layers:
            renpy.scene(l)

    config.scene = _scene
    _camera_viewer = CameraViewer()
    _transform_viewer = TransformViewer()

# add
init -1098 python:
    # overwrite keymap
    km = renpy.Keymap(
        rollback = renpy.rollback,
        screenshot = _screenshot,
        toggle_fullscreen = renpy.toggle_fullscreen,
        toggle_skip = _keymap_toggle_skipping,
        fast_skip = _fast_skip,
        game_menu = _invoke_game_menu,
        hide_windows = renpy.curried_call_in_new_context("_hide_windows"),
        launch_editor = _launch_editor,
        reload_game = _reload_game,
        developer = _developer,
        quit = renpy.quit_event,
        iconify = renpy.iconify,
        help = _help,
        choose_renderer = renpy.curried_call_in_new_context("_choose_renderer"),
        console = _console.enter,
        profile_once = _profile_once,
        self_voicing = Preference("self voicing", "toggle"),
        # variable_viewer = _dynamic_watch,
        camera_viewer = CameraViewer().viewer,
        variable_viewer = _dynamic_watch,
        transform_viewer = TransformViewer().viewer,
        )

    config.underlay = [ km ]

    del km


init 1100 python:
    config.locked = False
    config.keymap["camera_viewer"] = ['C']
    config.keymap["transform_viewer"] = ['T']
    config.keymap["variable_viewer"] = ['W']

    config.layers.append('debug')
    config.top_layers.append('debug')
    config.watcher_width = 600
    config.watcher_height = 300
    config.locked = True

    def put_clipboard(string):
        """
        :doc: other

        Put a string to clipboard
        """
        from pygame import scrap, locals
        scrap.init()
        scrap.put(locals.SCRAP_TEXT, string)

    def get_clipboard():
        """
        :doc: other

        Get a string from clipboard
        """
        from pygame import scrap, locals
        scrap.init()
        return scrap.get(locals.SCRAP_TEXT)
    renpy.get_clipboard = get_clipboard
    renpy.put_clipboard = put_clipboard
